' A Slightly Larger but Faster Parser (PB Assembler)
' enhanced version with look up table

' including demo with performance test

' Charles E V Pegge
' 5 August 2006
' charles@pegge.net
' www.pegge.net


' this parser is suitable for use in a script engine or for parsing any data strings
' that use symbols as well as words. Call it repeatedly to index sequential words
' and symbols in a string of text.

' I have written it to accelerate my scripting language; the machine code is only 124 bytes, but
' also uses a lookup table of 256 bytes. By reducing the branching logic, this runs about 10%
' faster than the previous non tabular version (in excess of 21 million words/sec on my Athlon 3200).

' The table gives added ease of configuration and flexibility. The upper part of the table
' may be dispensed with if the your source text uses 7 bit ascii only.

' General parsing rules:
'
' space, and all lower ascii and all non alphanumerics are treated as delimiters. eg:
' But an underscore _ (ascii 95) is treated as an alphanumeric eg:
'
'   FREE.LUNCH is interpreted as three words
'   FREE_LUNCH is one word

' all other non alphanumerics including '-' (minus) are self delimiting eg:
'
'   + - . ,  { } are always recognised as individual words even when unspaced
'
'
' But quoted text is treated as an entire word including the outer quotes.
' inner quotes can be embedded eg:
'
'    "She said 'HI' "
'    'She said "HI" '
'
' this code uses the EAX ECX EDX EBX EBP and ESI registers  -
' with minimal use of memory, most of the work is done in registers.
'
' you can use it inline or in its own subroutine or function.
'


' quick Ascii code reference:

'  various control chars 0-31:
'  null      0
'  htab      9
'  linefeed  10
'  vtab      11
'  formfeed  12
'  creturn   13
'  escape    27
'  space     32
'  !"#$%&'   33-39
'  ()*+,-./  40-47
'  0-9       48-57
'  :;<=>?@   58-64
'  A-Z       65-90
'  [\]^_`    91-96
'  a-z       97-122
'  {|}~      123-126



#COMPILE EXE
#DIM ALL
#INCLUDE "WIN32API.INC"

FUNCTION PBMAIN () AS LONG

' input/output variables

DIM az  AS STRING   ' Test string
DIM bwi AS DWORD    ' character position of word in the string
DIM lbi AS DWORD    ' length of string
DIM ii  AS DWORD    ' next position after the end of the word
DIM ip  AS BYTE PTR ' absolute address of the string

' test variables

DIM result    AS STRING
DIM testloops AS DWORD
DIM timing    AS DWORD
DIM wordcount AS DWORD

' Lookup Table

DIM tbl AS STRING
'0123456789ABCDEF'     '   codings
tbl= _                 '
"0000000000000000" + _ '0  LOWER ASCII
"0000000000000000" + _ '1  space/ignore    0
"0121111211111111" + _ '2  symbols         1 general 2 quote marks
"5555555555111111" + _ '3  numbers         5
"1666666666666666" + _ '4  alpha uppercase 6
"6666666666611118" + _ '5  alpha uppercase 6
"1777777777777777" + _ '6  alpha lowercase 7
"7777777777711111" + _ '7  alpha lowercase 7
"0000000000000000" + _ '8  HIGHER ASCII
"0000000000000000" + _ '9
"0000000000000000" + _ 'A
"0000000000000000" + _ 'B
"0000000000000000" + _ 'C
"0000000000000000" + _ 'D
"0000000000000000" + _ 'E
"0000000000000000" + _ 'F
""

' other codings available:
' self delimiting types: (like sybols)        3 4
' other types            (like alphanumerics) 8 9 eg underscore



DIM tblp AS BYTE PTR
tblp=STRPTR(tbl)

az="Old_wine    Free.Meal(Lunch,12.34)  "+CHR$(34)+"bon mot (7)"+CHR$(34)+"  "+CHR$(39)+CHR$(34)+"okay"+CHR$(34)+CHR$(39)

'measure speed

lbi=LEN(az)
ip=STRPTR(az)-1
timing=GetTickCount() ' ticks every millisecond
FOR testloops=1 TO 1000000 ' 1 million
 ii=1
 WHILE ii=<lbi
  GOSUB zbword
 WEND
NEXT
timing=GetTickCount()-Timing


'show results

result=az+":"+$CRLF
ii=1
WHILE ii=<lbi
 INCR wordcount
 GOSUB zbword
 result=result+MID$(az,bwi,ii-bwi)+$CRLF
WEND
MSGBOX result+$CRLF _
 +"words/millisec "+STR$(INT(1000000*wordcount/timing))+$CRLF _
 +"Length of code:"+STR$(CODEPTR(zeword)-CODEPTR(zbword))

' file dump hex
' DIM i AS DWORD, j AS DWORD
' DIM s AS STRING
' OPEN "code.txt" FOR OUTPUT AS 1
' i=CODEPTR(zbword)
' WHILE i<CODEPTR(zeword)
'  s=s+" "+RIGHT$("0"+HEX$(PEEK(i)),2)
'  INCR j: IF j=16 THEN j=0:PRINT#1,s:s=""
'  INCR i
' WEND
' PRINT#1,s 'flush
' CLOSE 1


EXIT FUNCTION

' the code itself

zbword:

! push esi
! push ebx
! push ebp                  ' save original base pointer
! mov edx,ip                ' pointer to start of string-1
! push edx                  ' hold this pointer on the stack
! push tblp                 ' hold the tbl pointer on the stack until all other variable
                            ' have been loaded into registers
! mov eax,edx               ' to find end of string
! add edx,lbi               ' points to the last character in the string
! add eax,ii                ' points to the character indexed by ii
! mov ecx,eax               ' default value for the start of word pointer
! dec eax                   ' prepare for pre incrementing loop
                            ' with basic variables installed,
                            ' it is now safe to overwrite the base pointer register
! mov ebp,eax               ' move the char pointer to the base pointer
! xor eax,eax               ' clear eax for another use
! pop ebx                   ' pop the tbl pointer into ebx ready to do lookups


next_space:                 ' for skipping leading spaces tabs etc
! inc ebp                   ' advance the char pointer
! cmp edx,ebp               ' check if end of string
! jl short start_of_word    ' exit loop if end of string
! mov al,[ebp]              ' get character
! cmp al,32                 ' is it a space or less ?
! jle short next_space      ' if so then try text character

start_of_word:              ' first non space char encountered

! mov ecx,ebp               ' remember start position

! cmp edx,ebp               ' check if end of string
! jl short end_of           ' exit if end of string

! mov esi,eax               ' use the char as an index
! mov al,[ebx+esi]          ' to look up its type code in tbl
! cmp al,&h34               ' if it is more than a symbol
! jg short next_letter      ' then it is an alphanumeric so try next char
! cmp al,&h32               ' is it a quote symbol?
! jz quote_string           ' if so then process whole quote
! inc bp                    ' else increment char pointer and
! jmp short end_of          ' return a self-delimiring symbol

quote_string:               ' this is a quoted item, we need to find the end quote mark

quote_string_traverse:      ' look for a matching quote mark
! mov al,[ebp]              ' reload quote mark
! inc ebp                   ' advance pointer for next char
! cmp edx,ebp               ' check if past the end of string
! jl short end_of           ' finish if so
! cmp al,[ebp]              ' check for matching quote mark
! jnz short quote_string_traverse
                            ' if no match then try next
! inc ebp                   ' advance past end of quote
! jmp short end_of          ' we have our quote so finish

next_letter:                ' scanning an alpha numeric. Could be delimited by a symbol, space or lesser ascii

! inc ebp                   ' advance pointer
! cmp edx,ebp               ' past end of string?
! jl short end_of           ' if so then finish
! mov al,[ebp]              ' get char
! mov esi,eax               ' use the char as an index
! mov al,[ebx+esi]          ' then lookup its type in tbl
! cmp al,&h34               ' if it is more than a space or symbol
! jg short next_letter      ' then it is an alphanumeric so try then next char

end_of:                     ' results are ready to pass back to BASIC variables

! pop edx                   ' pop original position of string pointer
! mov eax,ebp               ' hold the end position in eax before we pop the bp register
! pop ebp                   ' restore base pointer register
! pop ebx                   ' and the base register
! pop esi                   ' and the si index register
                            ' so it is now safe to store to BASIC variables
! sub ecx,edx               ' derive our start of word index
! mov bwi,ecx               ' store result in bwi
! sub eax,edx               ' derive the after-word index
! mov ii,eax                ' assign value to ii

                            ' if the string is null or contains no words or symbols
                            ' then both bwi and ii will be greater than the length of the string

zeword:                    ' this label is just for measuring the length of the assembled code

RETURN

END FUNCTION
